﻿using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Mvc;
using Newtonsoft.Json.Linq;
using System.Diagnostics;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Threading.Tasks;
using TestClient.Models;

namespace TestClient.Controllers
{
    /// <summary>
    /// 홈 컨트롤러
    /// </summary>
    public class HomeController : Controller
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 인덱스 페이지 처리하기 - Index()

        /// <summary>
        /// 인덱스 페이지 처리하기
        /// </summary>
        /// <returns>액션 결과</returns>
        public IActionResult Index()
        {
            return View();
        }

        #endregion
        #region 개인 정보 보호 정책 페이지 처리하기 - Privacy()

        /// <summary>
        /// 개인 정보 보호 정책 페이지 처리하기
        /// </summary>
        /// <returns>액션 결과</returns>
        public IActionResult Privacy()
        {
            return View();
        }

        #endregion
        #region 에러 페이지 처리하기 - Error()

        /// <summary>
        /// 에러 페이지 처리하기
        /// </summary>
        /// <returns>액션 결과</returns>
        [ResponseCache(Duration = 0, Location = ResponseCacheLocation.None, NoStore = true)]
        public IActionResult Error()
        {
            return View(new ErrorViewModel { RequestID = Activity.Current?.Id ?? HttpContext.TraceIdentifier });
        }

        #endregion
        #region 로그아웃 페이지 처리하기 - Logout()

        /// <summary>
        /// 로그아웃 페이지 처리하기
        /// </summary>
        /// <returns>액션 결과</returns>
        public IActionResult Logout()
        {
            return SignOut("Cookies", "oidc");
        }

        #endregion
        #region API 호출 페이지 처리하기 - CallAPI()

        /// <summary>
        /// API 호출 페이지 처리하기
        /// </summary>
        /// <returns>액션 결과 태스크</returns>
        public async Task<IActionResult> CallAPI()
        {
            string accessToken = await HttpContext.GetTokenAsync("access_token");

            HttpClient client = new HttpClient();

            client.DefaultRequestHeaders.Authorization = new AuthenticationHeaderValue("Bearer", accessToken);

            string content = await client.GetStringAsync("https://localhost:44310/identity/index");

            ViewData["JSON"] = JArray.Parse(content).ToString();

            return View();
        }

        #endregion
    }
}