using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using System.IdentityModel.Tokens.Jwt;

namespace TestClient
{
    /// <summary>
    /// 
    /// </summary>
    public class Startup
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  ÷ ϱ - ConfigureServices(services)

        /// <summary>
        ///  ÷ ϱ
        /// </summary>
        /// <param name="services"> ÷</param>
        public void ConfigureServices(IServiceCollection services)
        {
            JwtSecurityTokenHandler.DefaultMapInboundClaims = false;

            services.AddAuthentication
            (
                options =>
                {
                    options.DefaultScheme          = "Cookies";
                    options.DefaultChallengeScheme = "oidc";
                }
            )
            .AddCookie("Cookies")
            .AddOpenIdConnect
            (
                "oidc",
                options =>
                {
                    options.Authority                     = "https://localhost:44300";
                    options.ClientId                      = "CLIENTID0002";
                    options.ClientSecret                  = "CLIENTSECRET0002";
                    options.ResponseType                  = "code";
                    options.SaveTokens                    = true;
                    options.GetClaimsFromUserInfoEndpoint = true;

                    options.Scope.Add("profile"       );
                    options.Scope.Add("API1"          );
                    options.Scope.Add("offline_access");
                }
            );

            services.AddControllersWithViews();
        }

        #endregion
        #region ϱ - Configure(app, environment)

        /// <summary>
        /// ϱ
        /// </summary>
        /// <param name="app">ø̼ </param>
        /// <param name="environment"> ȣƮ ȯ</param>
        public void Configure(IApplicationBuilder app, IWebHostEnvironment environment)
        {
            if(environment.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }
            else
            {
                app.UseExceptionHandler("/Home/Error");

                app.UseHsts();
            }

            app.UseHttpsRedirection();

            app.UseStaticFiles();

            app.UseRouting();

            app.UseAuthentication();

            app.UseAuthorization();

            app.UseEndpoints
            (
                endpoints =>
                {
                    endpoints.MapDefaultControllerRoute().RequireAuthorization();
                }
            );
        }

        #endregion
    }
}