﻿using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.Filters;

namespace TestIdentityServer.Attributes
{
    /// <summary>
    /// 보안 헤더 어트리뷰트
    /// </summary>
    public class SecurityHeadersAttribute : ActionFilterAttribute
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 결과 실행시 처리하기 - OnResultExecuting(context)

        /// <summary>
        /// 결과 실행시 처리하기
        /// </summary>
        /// <param name="context">결과 실행시 컨텍스트</param>
        public override void OnResultExecuting(ResultExecutingContext context)
        {
            IActionResult result = context.Result;

            if(result is ViewResult)
            {
                // https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Content-Type-Options
                if(!context.HttpContext.Response.Headers.ContainsKey("X-Content-Type-Options"))
                {
                    context.HttpContext.Response.Headers.Add("X-Content-Type-Options", "nosniff");
                }

                // https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Frame-Options
                if(!context.HttpContext.Response.Headers.ContainsKey("X-Frame-Options"))
                {
                    context.HttpContext.Response.Headers.Add("X-Frame-Options", "SAMEORIGIN");
                }

                // https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy
                string csp = "default-src 'self'; object-src 'none'; frame-ancestors 'none'; " + 
                             "sandbox allow-forms allow-same-origin allow-scripts; base-uri 'self';";

                // 또한 프로덕션을 위해 HTTPS를 확보한 후 업그레이드-비보안-요청 추가를 고려한다.
                // csp += "upgrade-insecure-requests;";
                // 또한 트위터에서 클라이언트 이미지를 표시해야 하는 경우의 예이다.
                // csp += "img-src 'self' https://pbs.twimg.com;";

                // 표준 호환 브라우저의 경우 실행한다.
                if(!context.HttpContext.Response.Headers.ContainsKey("Content-Security-Policy"))
                {
                    context.HttpContext.Response.Headers.Add("Content-Security-Policy", csp);
                }

                // IE인 경우 실행한다.
                if(!context.HttpContext.Response.Headers.ContainsKey("X-Content-Security-Policy"))
                {
                    context.HttpContext.Response.Headers.Add("X-Content-Security-Policy", csp);
                }

                // https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Referrer-Policy
                string referrerPolicy = "no-referrer";

                if(!context.HttpContext.Response.Headers.ContainsKey("Referrer-Policy"))
                {
                    context.HttpContext.Response.Headers.Add("Referrer-Policy", referrerPolicy);
                }
            }
        }

        #endregion
    }
}