﻿using System.Collections.Generic;

using IdentityServer4;
using IdentityServer4.Models;

namespace TestIdentityServer
{
    /// <summary>
    /// 구성
    /// </summary>
    public static class Configuration
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 신원 리소스 리스트 구하기 - GetIdentityResourceList()

        /// <summary>
        /// 신원 리소스 리스트 구하기
        /// </summary>
        /// <returns>신원 리소스 리스트</returns>
        public static List<IdentityResource> GetIdentityResourceList()
        {
            return new List<IdentityResource>
            { 
                new IdentityResources.OpenId(),
                new IdentityResources.Profile()
            };
        }

        #endregion
        #region API 범위 리스트 구하기 - GetAPIScopeList()

        /// <summary>
        /// API 범위 리스트 구하기
        /// </summary>
        /// <returns>API 범위 리스트</returns>
        public static IEnumerable<ApiScope> GetAPIScopeList()
        {
            return new List<ApiScope>
            {
                new ApiScope("API1", "API 1")
            };
        }

        #endregion
        #region 클라이언트 리스트 구하기 - GetClientList()

        /// <summary>
        /// 클라이언트 리스트 구하기
        /// </summary>
        /// <return>클라이언트 리스트</return>
        public static IEnumerable<Client> GetClientList()
        {
            return new List<Client>
            {
                new Client
                {
                    AllowedGrantTypes = GrantTypes.ClientCredentials,
                    ClientId          = "CLIENTID0001",
                    ClientSecrets     = { new Secret("CLIENTSECRET0001".Sha256()) },
                    AllowedScopes     = { "API1" }
                },
                new Client
                {
                    AllowedGrantTypes      = GrantTypes.Code,
                    ClientId               = "CLIENTID0002",
                    ClientSecrets          = { new Secret("CLIENTSECRET0002".Sha256()) },
                    RedirectUris           = { "https://localhost:44320/signin-oidc" },
                    PostLogoutRedirectUris = { "https://localhost:44320/signout-callback-oidc" },
                    AllowOfflineAccess     = true,
                    AllowedScopes          = new List<string>
                    {
                        IdentityServerConstants.StandardScopes.OpenId,
                        IdentityServerConstants.StandardScopes.Profile,
                        "API1"
                    }
                }
            };
        }

        #endregion
    }
}