using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

using IdentityModel;
using IdentityServer4;
using IdentityServer4.Events;
using IdentityServer4.Extensions;
using IdentityServer4.Models;
using IdentityServer4.Services;
using IdentityServer4.Stores;
using IdentityServer4.Test;

using TestIdentityServer.Attributes;
using TestIdentityServer.Extensions;
using TestIdentityServer.Models;

namespace TestIdentityServer
{
    /// <summary>
    ///  Ʈѷ
    /// </summary>
    [SecurityHeaders]
    [AllowAnonymous]
    public class AccountController : Controller
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// ſ  ȭ 
        /// </summary>
        private readonly IIdentityServerInteractionService identityServerInteractionService;

        /// <summary>
        /// Ŭ̾Ʈ 
        /// </summary>
        private readonly IClientStore clientStore;

        /// <summary>
        ///  ȹ 
        /// </summary>
        private readonly IAuthenticationSchemeProvider authenticationSchemeProvider;

        /// <summary>
        /// ̺Ʈ 
        /// </summary>
        private readonly IEventService eventService;

        /// <summary>
        /// ׽Ʈ  
        /// </summary>
        private readonly TestUserStore testUserStore;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - AccountController(identityServerInteractionService, clientStore, authenticationSchemeProvider, eventService, testUserStore)

        /// <summary>
        /// 
        /// </summary>
        /// <param name="identityServerInteractionService">ſ  ȭ </param>
        /// <param name="clientStore">Ŭ̾Ʈ </param>
        /// <param name="authenticationSchemeProvider"> ȹ </param>
        /// <param name="eventService">̺Ʈ </param>
        /// <param name="testUserStore">׽Ʈ  </param>
        public AccountController
        (
            IIdentityServerInteractionService identityServerInteractionService,
            IClientStore                      clientStore,
            IAuthenticationSchemeProvider     authenticationSchemeProvider,
            IEventService                     eventService,
            TestUserStore                     testUserStore = null
        )
        {
            this.identityServerInteractionService = identityServerInteractionService;
            this.clientStore                      = clientStore;
            this.authenticationSchemeProvider     = authenticationSchemeProvider;
            this.eventService                     = eventService;
            this.testUserStore                    = testUserStore ?? new TestUserStore(TestUserData.UserList);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region α  óϱ - Login(returnURL)

        /// <summary>
        /// α  óϱ
        /// </summary>
        /// <param name="returnURL">ȯ URL</param>
        /// <returns>׼  ½ũ</returns>
        [HttpGet]
        public async Task<IActionResult> Login(string returnURL)
        {
            // α  ǥ    ֵ  ۼѴ.
            LoginViewModel vm = await BuildLoginViewModelAsync(returnURL);

            if(vm.IsExternalLoginOnly)
            {
                // α ɼ ϳ ̸ ܺ ̴.
                return RedirectToAction("Challenge", "External", new { scheme = vm.ExternalLoginScheme, returnURL });
            }

            return View(vm);
        }

        #endregion
        #region α  óϱ - (model, button)

        /// <summary>
        /// α  óϱ
        /// </summary>
        /// <param name="model">α Է </param>
        /// <param name="button">ư Ÿ</param>
        /// <returns>׼  ½ũ</returns>
        [HttpPost]
        [ValidateAntiForgeryToken]
        public async Task<IActionResult> Login(LoginInputModel model, string button)
        {
            //  û ؽƮ ִ ȮѴ.
            AuthorizationRequest authorizationRequest = await this.identityServerInteractionService.GetAuthorizationContextAsync
            (
                model.ReturnURL
            );

            // ڰ "" ư Ŭ  óѴ.
            if(button != "login")
            {
                if(authorizationRequest != null)
                {
                    // ڰ ϸ Ǹ ź ó  IdentityServer ٽ 
                    // ( Ŭ̾Ʈ ǰ ʿ  쿡 ش).
                    // ׷ ׼ ź OIDC   Ŭ̾Ʈ ٽ ۵ȴ. 
                    await this.identityServerInteractionService.DenyAuthorizationAsync
                    (
                        authorizationRequest,
                        AuthorizationError.AccessDenied
                    );

                    // GetAuthorizationContextAsync null ƴ  ȯ߱  model.ReturnUrl ŷ  ִ.
                    if(authorizationRequest.IsNativeClient())
                    {
                        // Ŭ̾Ʈ Ƽ̹Ƿ  ȯϴ  ̷ 
                        //  ڿ   UX ϱ  ̴.
                        return this.LoadingPage("Redirect", model.ReturnURL);
                    }

                    return Redirect(model.ReturnURL);
                }
                else
                {
                    // ùٸ ؽƮ   Ȩ  ư.
                    return Redirect("~/");
                }
            }

            if(ModelState.IsValid)
            {
                // ޸  ҿ   ̸/ȣ Ѵ.
                if(this.testUserStore.ValidateCredentials(model.UserName, model.Password))
                {
                    TestUser testUser = this.testUserStore.FindByUsername(model.UserName);

                    await this.eventService.RaiseAsync
                    (
                        new UserLoginSuccessEvent
                        (
                            testUser.Username,
                            testUser.SubjectId,
                            testUser.Username,
                            clientId : authorizationRequest?.Client.ClientId
                        )
                    );

                    // ڰ "ϱ"  쿡 ⿡  Ḧ Ѵ.
                    // ׷  Ű ̵  ῡ Ѵ.
                    AuthenticationProperties authenticationProperties = null;

                    if(AccountOptions.AllowRememberLogin && model.RememberLogin)
                    {
                        authenticationProperties = new AuthenticationProperties
                        {
                            IsPersistent = true,
                            ExpiresUtc   = DateTimeOffset.UtcNow.Add(AccountOptions.RememberLoginDuration)
                        };
                    };

                    // Ʈ ID   ̸  Ű ߱Ѵ.
                    IdentityServerUser identityServerUser = new IdentityServerUser(testUser.SubjectId)
                    {
                        DisplayName = testUser.Username
                    };

                    await HttpContext.SignInAsync(identityServerUser, authenticationProperties);

                    if(authorizationRequest != null)
                    {
                        if(authorizationRequest.IsNativeClient())
                        {
                            // Ŭ̾Ʈ ⺻̹Ƿ  ȯϴ  ̷ 
                            //  ڿ   UX ϱ  ̴.
                            return this.LoadingPage("Redirect", model.ReturnURL);
                        }

                        // GetAuthorizationContextAsync null ƴ  ȯ߱  model.ReturnUrl ŷ  ִ.
                        return Redirect(model.ReturnURL);
                    }

                    //   ûѴ.
                    if(Url.IsLocalUrl(model.ReturnURL))
                    {
                        return Redirect(model.ReturnURL);
                    }
                    else if(string.IsNullOrEmpty(model.ReturnURL))
                    {
                        return Redirect("~/");
                    }
                    else
                    {
                        // ڰ Ǽ ũ Ŭ  ִ - ϵǾ Ѵ.
                        throw new Exception("invalid return URL");
                    }
                }

                await this.eventService.RaiseAsync
                (
                    new UserLoginFailureEvent
                    (
                        model.UserName,
                        "invalid credentials",
                        clientId : authorizationRequest?.Client.ClientId
                    )
                );

                ModelState.AddModelError(string.Empty, AccountOptions.InvalidCredentialsErrorMessage);
            }

            //  ߻ߴ.  ִ  ǥѴ.
            LoginViewModel vm = await BuildLoginViewModelAsync(model);

            return View(vm);
        }

        #endregion
        #region α׾ƿ  óϱ - Logout(logoutID)

        /// <summary>
        /// α׾ƿ  óϱ
        /// </summary>
        /// <param name="logoutID">α׾ƿ ID</param>
        /// <returns>׼  ½ũ</returns>
        [HttpGet]
        public async Task<IActionResult> Logout(string logoutID)
        {
            // α ƿ  ǥ    ֵ α׾ƿ   .
            LogoutViewModel vm = await BuildLogoutViewModelAsync(logoutID);

            if(vm.ShowLogoutPrompt == false)
            {
                // α ƿ û IdentityServer   
                // Ʈ ǥ ʿ䰡  ڸ  α׾ƿ   ִ.
                return await Logout(vm);
            }

            return View(vm);
        }

        #endregion
        #region α׾ƿ  óϱ - Logout(model)

        /// <summary>
        /// α׾ƿ  óϱ
        /// </summary>
        /// <param name="model">α׾ƿ Է </param>
        /// <returns>׼  ½ũ</returns>
        [HttpPost]
        [ValidateAntiForgeryToken]
        public async Task<IActionResult> Logout(LogoutInputModel model)
        {
            // α׾ƿ  ǥ    ֵ α׾ƿ Ϸ   .
            LoggedOutViewModel vm = await BuildLoggedOutViewModelAsync(model.LogoutID);

            if(User?.Identity.IsAuthenticated == true)
            {
                //   Ű Ѵ.
                await HttpContext.SignOutAsync();

                // α׾ƿ ̺Ʈ ߻Ų.
                await this.eventService.RaiseAsync(new UserLogoutSuccessEvent(User.GetSubjectId(), User.GetDisplayName()));
            }

            // Ʈ ID ڿ α׾ƿ Ʈؾ ϴ ȮѴ.
            if(vm.TriggerExternalLogout)
            {
                // ڰ α׾ƿ  Ʈ ڰ ٽ  𷺼ǵǵ ȯ URL ۼѴ.
                // ̸  ̱  ƿ ó Ϸ  ִ.
                string url = Url.Action("Logout", new { logoutID = vm.LogoutID });

                // ̷ ϸ α׾ƿ  ܺ ڷ 𷺼 Ʈŵȴ.
                return SignOut(new AuthenticationProperties { RedirectUri = url }, vm.ExternalAuthenticationScheme);
            }

            return View("LoggedOut", vm);
        }

        #endregion
        #region ׼ źν  óϱ - AccessDenied()

        /// <summary>
        /// ׼ źν  óϱ
        /// </summary>
        /// <returns>׼ </returns>
        [HttpGet]
        public IActionResult AccessDenied()
        {
            return View();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region α    (񵿱) - BuildLoginViewModelAsync(returnURL)

        /// <summary>
        /// α    (񵿱)
        /// </summary>
        /// <param name="returnURL">ȯ URL</param>
        /// <returns>α   ½ũ</returns>
        private async Task<LoginViewModel> BuildLoginViewModelAsync(string returnURL)
        {
            AuthorizationRequest authorizationRequest = await this.identityServerInteractionService.GetAuthorizationContextAsync
            (
                returnURL
            );

            if
            (
                authorizationRequest?.IdP != null &&
                await this.authenticationSchemeProvider.GetSchemeAsync(authorizationRequest.IdP) != null
            )
            {
                bool isLocalIdentityProvider = authorizationRequest.IdP == IdentityServerConstants.LocalIdentityProvider;

                // ̴ UI ܶŰ ϳ ܺ IdP Ʈϱ  ̴.
                LoginViewModel vm = new LoginViewModel
                {
                    EnableLocalLogin = isLocalIdentityProvider,
                    ReturnURL        = returnURL,
                    UserName         = authorizationRequest?.LoginHint,
                };

                if(!isLocalIdentityProvider)
                {
                    vm.ExternalProviderEnumerable = new[]
                    {
                        new ExternalProvider { AuthenticationScheme = authorizationRequest.IdP }
                    };
                }

                return vm;
            }

            IEnumerable<AuthenticationScheme> authenticationSchemeEnumerable = await this.authenticationSchemeProvider.GetAllSchemesAsync();

            List<ExternalProvider> externalProviderList = authenticationSchemeEnumerable
                .Where(x => x.DisplayName != null)
                .Select
                (
                    x => new ExternalProvider
                    {
                        DisplayName          = x.DisplayName ?? x.Name,
                        AuthenticationScheme = x.Name
                    }
                )
                .ToList();

            bool allowLocal = true;

            if(authorizationRequest?.Client.ClientId != null)
            {
                Client client = await this.clientStore.FindEnabledClientByIdAsync(authorizationRequest.Client.ClientId);

                if(client != null)
                {
                    allowLocal = client.EnableLocalLogin;

                    if(client.IdentityProviderRestrictions != null && client.IdentityProviderRestrictions.Any())
                    {
                        externalProviderList = externalProviderList.Where
                        (
                            provider => client.IdentityProviderRestrictions.Contains(provider.AuthenticationScheme)
                        )
                        .ToList();
                    }
                }
            }

            return new LoginViewModel
            {
                AllowRememberLogin         = AccountOptions.AllowRememberLogin,
                EnableLocalLogin           = allowLocal && AccountOptions.AllowLocalLogin,
                ReturnURL                  = returnURL,
                UserName                   = authorizationRequest?.LoginHint,
                ExternalProviderEnumerable = externalProviderList.ToArray()
            };
        }

        #endregion
        #region α    (񵿱) - BuildLoginViewModelAsync(model)

        /// <summary>
        /// α    (񵿱)
        /// </summary>
        /// <param name="model">α Է </param>
        /// <returns>α   ½ũ</returns>
        private async Task<LoginViewModel> BuildLoginViewModelAsync(LoginInputModel model)
        {
            LoginViewModel vm = await BuildLoginViewModelAsync(model.ReturnURL);

            vm.UserName      = model.UserName;
            vm.RememberLogin = model.RememberLogin;

            return vm;
        }

        #endregion
        #region α׾ƿ    (񵿱) - BuildLogoutViewModelAsync(logoutID)

        /// <summary>
        /// α׾ƿ    (񵿱)
        /// </summary>
        /// <param name="logoutID">α׾ƿ ID</param>
        /// <returns>α׾ƿ   ½ũ</returns>
        private async Task<LogoutViewModel> BuildLogoutViewModelAsync(string logoutID)
        {
            LogoutViewModel vm = new LogoutViewModel
            {
                LogoutID         = logoutID,
                ShowLogoutPrompt = AccountOptions.ShowLogoutPrompt
            };

            if(User?.Identity.IsAuthenticated != true)
            {
                // ڰ    α׾ƿ  ǥѴ.
                vm.ShowLogoutPrompt = false;

                return vm;
            }

            LogoutRequest logoutRequest = await this.identityServerInteractionService.GetLogoutContextAsync(logoutID);

            if(logoutRequest?.ShowSignoutPrompt == false)
            {
                // ڵ α׾ƿص ϴ.
                vm.ShowLogoutPrompt = false;

                return vm;
            }

            // α׾ƿ Ʈ ǥѴ.
            // ̰ ڰ ٸ Ǽ    ڵ α׾ƿǴ  Ѵ.
            return vm;
        }

        #endregion
        #region α׾ƿ Ϸ    (񵿱) - BuildLoggedOutViewModelAsync(logoutID)

        /// <summary>
        /// α׾ƿ Ϸ    (񵿱)
        /// </summary>
        /// <param name="logoutID">α׾ƿ ID</param>
        /// <returns>α׾ƿ Ϸ   ½ũ</returns>
        private async Task<LoggedOutViewModel> BuildLoggedOutViewModelAsync(string logoutID)
        {
            // ؽƮ  ´(Ŭ̾Ʈ, Ʈ α׾ƿ  URI  ̼ǵ α׾ƿ  iframe).
            LogoutRequest logoutRequest = await this.identityServerInteractionService.GetLogoutContextAsync(logoutID);

            LoggedOutViewModel vm = new LoggedOutViewModel
            {
                AutomaticRedirectAfterLogOut = AccountOptions.AutomaticRedirectAfterLogout,
                PostLogoutRedirectURI        = logoutRequest?.PostLogoutRedirectUri,
                ClientName                   = string.IsNullOrEmpty(logoutRequest?.ClientName)
                                                   ? logoutRequest?.ClientId : logoutRequest?.ClientName,
                LogOutIframeURL              = logoutRequest?.SignOutIFrameUrl,
                LogoutID                     = logoutID
            };

            if(User?.Identity.IsAuthenticated == true)
            {
                string idp = User.FindFirst(JwtClaimTypes.IdentityProvider)?.Value;

                if(idp != null && idp != IdentityServerConstants.LocalIdentityProvider)
                {
                    bool providerSupportsLogout = await HttpContext.GetSchemeSupportsSignOutAsync(idp);

                    if(providerSupportsLogout)
                    {
                        if(vm.LogoutID == null)
                        {
                            //  α׾ƿ ؽƮ     Ѵ.
                            // ̷ ϸ α׾ƿϱ   α ڷκ ʿ  ĸóϰ
                            // α ƿ  ܺ IdP Ѵ.
                            vm.LogoutID = await this.identityServerInteractionService.CreateLogoutContextAsync();
                        }

                        vm.ExternalAuthenticationScheme = idp;
                    }
                }
            }

            return vm;
        }

        #endregion
    }
}