using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Claims;
using System.Threading.Tasks;

using IdentityModel;
using IdentityServer4;
using IdentityServer4.Events;
using IdentityServer4.Services;
using IdentityServer4.Stores;
using IdentityServer4.Test;
using IdentityServer4.Models;

using TestIdentityServer.Attributes;
using TestIdentityServer.Extensions;
using TestIdentityServer.Models;

namespace TestIdentityServer
{
    /// <summary>
    /// ܺ Ʈѷ
    /// </summary>
    [SecurityHeaders]
    [AllowAnonymous]
    public class ExternalController : Controller
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// ſ  ȭ 
        /// </summary>
        private readonly IIdentityServerInteractionService identityServerInteractionService;

        /// <summary>
        /// Ŭ̾Ʈ 
        /// </summary>
        private readonly IClientStore clientStore;

        /// <summary>
        /// α ϱ
        /// </summary>
        private readonly ILogger<ExternalController> logger;

        /// <summary>
        /// ̺Ʈ 
        /// </summary>
        private readonly IEventService eventService;

        /// <summary>
        /// ׽Ʈ  
        /// </summary>
        private readonly TestUserStore testUserStore;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - ExternalController(identityServerInteractionService, clientStore, eventService, logger, testUserStore)

        /// <summary>
        /// 
        /// </summary>
        /// <param name="identityServerInteractionService">ſ  ȭ </param>
        /// <param name="clientStore">Ŭ̾Ʈ </param>
        /// <param name="eventService">̺Ʈ </param>
        /// <param name="logger">α ϱ</param>
        /// <param name="testUserStore">׽Ʈ  </param>
        public ExternalController
        (
            IIdentityServerInteractionService identityServerInteractionService,
            IClientStore                      clientStore,
            IEventService                     eventService,
            ILogger<ExternalController>       logger,
            TestUserStore                     testUserStore = null
        )
        {
            this.identityServerInteractionService = identityServerInteractionService;
            this.clientStore                      = clientStore;
            this.eventService                     = eventService;
            this.logger                           = logger;
            // TestUserStore DI   ۷ι  ÷Ǹ Ѵ.
            // ⿡   ID  ̺귯( : ASP.NET ID)   ִ.
            this.testUserStore                    = testUserStore ?? new TestUserStore(TestUserData.UserList);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region   óϱ - Challenge(scheme, returnURL)

        /// <summary>
        ///   óϱ
        /// </summary>
        /// <param name="scheme">ȹ</param>
        /// <param name="returnURL">ȯ URL</param>
        /// <returns>׼ </returns>
        /// <remarks>ܺ  ڿ  պ ʱȭѴ.</remarks>
        [HttpGet]
        public IActionResult Challenge(string scheme, string returnURL)
        {
            if(string.IsNullOrEmpty(returnURL))
            {
                returnURL = "~/";
            }

            // returnURL ˻Ѵ - ȿ OIDC URL̰ų   ư.
            if(Url.IsLocalUrl(returnURL) == false && this.identityServerInteractionService.IsValidReturnUrl(returnURL) == false)
            {
                // ڰ Ǽ ũ Ŭ  ִ - ϵǾ Ѵ.
                throw new Exception("invalid return URL");
            }
            
            //  ϰ ȯ URL  ü踦 պѴ.
            AuthenticationProperties authenticationProperties = new AuthenticationProperties
            {
                RedirectUri = Url.Action(nameof(Callback)), 
                Items       =
                {
                    { "returnUrl", returnURL }, 
                    { "scheme"   , scheme    }
                }
            };

            return Challenge(authenticationProperties, scheme);
        }

        #endregion
        #region ݹ  óϱ - Callback()

        /// <summary>
        /// ݹ  óϱ
        /// </summary>
        /// <returns>׼  ½ũ</returns>
        /// <remarks>ܺ   óѴ.</remarks>
        [HttpGet]
        public async Task<IActionResult> Callback()
        {
            // ӽ Ű ܺ ID д´.
            AuthenticateResult authenticateResult = await HttpContext.AuthenticateAsync
            (
                IdentityServerConstants.ExternalCookieAuthenticationScheme
            );

            if(authenticateResult?.Succeeded != true)
            {
                throw new Exception("External authentication error");
            }

            if(this.logger.IsEnabled(LogLevel.Debug))
            {
                IEnumerable<string> externalClaimEnumberable = authenticateResult.Principal.Claims.Select
                (
                    claim => $"{claim.Type} : {claim.Value}"
                );

                this.logger.LogDebug("External claims : {@claims}", externalClaimEnumberable);
            }

            //   ܺ   ȸѴ.
            var (testUser, provider, providerUserID, claimEnumerable) = FindUserFromExternalProvider(authenticateResult);

            if(testUser == null)
            {
                //  ÿ      ũ ÷θ   ִ.
                //  ÿ    ʴ´.
                // ο ܺ ڸ ڵ κϱ⸸ ϸ ȴ.
                testUser = AutoProvisionUser(provider, providerUserID, claimEnumerable);
            }

            // ̸   Ư ݿ  ߰ Ŭ Ǵ Ӽ ϰ   Ű   ִ.
            // Ϲ ش ݿ α׾ƿϴ  ʿ ͸ ϴ  ȴ.
            List<Claim> additionalLocalClaimList = new List<Claim>();

            AuthenticationProperties localLoginAuthenticationProperties = new AuthenticationProperties();

            ProcessLoginCallback(authenticateResult, additionalLocalClaimList, localLoginAuthenticationProperties);
            
            // ڿ   Ű ߱Ѵ.
            IdentityServerUser identityServerUser = new IdentityServerUser(testUser.SubjectId)
            {
                DisplayName      = testUser.Username,
                IdentityProvider = provider,
                AdditionalClaims = additionalLocalClaimList
            };

            await HttpContext.SignInAsync(identityServerUser, localLoginAuthenticationProperties);

            // ܺ  Ǵ ӽ Ű Ѵ.
            await HttpContext.SignOutAsync(IdentityServerConstants.ExternalCookieAuthenticationScheme);

            // ȯ URL Ѵ.
            string returnURL = authenticateResult.Properties.Items["returnUrl"] ?? "~/";

            // ܺ α OIDC û ؽƮ ִ ȮѴ.
            AuthorizationRequest authorizationRequest = await this.identityServerInteractionService.GetAuthorizationContextAsync(returnURL);

            await this.eventService.RaiseAsync
            (
                new UserLoginSuccessEvent
                (
                    provider,
                    providerUserID,
                    testUser.SubjectId,
                    testUser.Username,
                    true,
                    authorizationRequest?.Client.ClientId
                )
            );

            if(authorizationRequest != null)
            {
                if(authorizationRequest.IsNativeClient())
                {
                    // Ŭ̾Ʈ ⺻̹Ƿ  ȯϴ  ̷   ڿ   UX ϱ  ̴.
                    return this.LoadingPage("Redirect", returnURL);
                }
            }

            return Redirect(returnURL);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region ܺ ڿ  ã - FindUserFromExternalProvider(authenticateResult)

        /// <summary>
        /// ܺ ڿ  ã
        /// </summary>
        /// <param name="authenticateResult"> </param>
        /// <returns>(׽Ʈ , ,   ID, Ŭ  )</returns>
        private (TestUser testUser, string provider, string providerUserID, IEnumerable<Claim> claimEnumerable) FindUserFromExternalProvider
        (
            AuthenticateResult authenticateResult
        )
        {
            ClaimsPrincipal externalUserClaimsPrincipal = authenticateResult.Principal;

            // ڰ ߱ ܺ   ID ȮѴ.
            //  Ϲ Ŭ  ܺ ڿ  sub Ŭ  NameIdentifier̸ ٸ Ŭ    ִ.
            Claim userIDClaim = externalUserClaimsPrincipal.FindFirst(JwtClaimTypes.Subject) ??
                              externalUserClaimsPrincipal.FindFirst(ClaimTypes.NameIdentifier) ??
                              throw new Exception("Unknown userid");

            // ڸ κ  ߰ Ŭ  ʵ  ID Ŭ Ѵ.
            List<Claim> claimList = externalUserClaimsPrincipal.Claims.ToList();

            claimList.Remove(userIDClaim);

            string provider       = authenticateResult.Properties.Items["scheme"];
            string providerUserID = userIDClaim.Value;

            // ܺ ڸ ã´.
            TestUser testUser = this.testUserStore.FindByExternalProvider(provider, providerUserID);

            return (testUser, provider, providerUserID, claimList);
        }

        #endregion
        #region  ڵ κϱ - AutoProvisionUser(provider, providerUserID, claimEnumerable)

        /// <summary>
        ///  ڵ κϱ
        /// </summary>
        /// <param name="provider"></param>
        /// <param name="providerUserID">  ID</param>
        /// <param name="claimEnumerable">Ŭ  </param>
        /// <returns>׽Ʈ </returns>
        private TestUser AutoProvisionUser(string provider, string providerUserID, IEnumerable<Claim> claimEnumerable)
        {
            TestUser testUser = this.testUserStore.AutoProvisionUser(provider, providerUserID, claimEnumerable.ToList());

            return testUser;
        }

        #endregion
        #region α ݹ óϱ - ProcessLoginCallback(externalAuthenticateResult, localClaimList, localLoginAuthenticationProperties)

        /// <summary>
        /// α ݹ óϱ
        /// </summary>
        /// <param name="externalAuthenticateResult">ܺ  </param>
        /// <param name="localClaimList"> Ŭ Ʈ</param>
        /// <param name="localLoginAuthenticationProperties"> α  Ӽ</param>
        /// <remarks>
        ///  ܺ α OIDC   α׾ƿ ۾  ؾ  Ư  ִ.
        /// ̰ WS-Fed, SAML2p Ǵ Ÿ ݿ  ٸ.
        /// </remarks>
        private void ProcessLoginCallback
        (
            AuthenticateResult       externalAuthenticateResult,
            List<Claim>              localClaimList,
            AuthenticationProperties localLoginAuthenticationProperties
        )
        {
            // ܺ ý  ID Ŭ     ƿ   ֵ Ѵ.
            Claim sessionIDClaim = externalAuthenticateResult.Principal.Claims.FirstOrDefault(x => x.Type == JwtClaimTypes.SessionId);

            if(sessionIDClaim != null)
            {
                localClaimList.Add(new Claim(JwtClaimTypes.SessionId, sessionIDClaim.Value));
            }

            // ܺ ڰ id_token   α ƿ  Ѵ.
            string idToken = externalAuthenticateResult.Properties.GetTokenValue("id_token");

            if(idToken != null)
            {
                localLoginAuthenticationProperties.StoreTokens(new[] { new AuthenticationToken { Name = "id_token", Value = idToken } });
            }
        }

        #endregion
    }
}